unit Basic;

interface

uses SysUtils, Windows, Messages, Classes, ComCtrls, SyncObjs, StrUtils;


type
  TSafeStringList = class(TStringList)
  private
    CriticalSection: TCriticalSection;
  public
    procedure Lock;
    procedure Unlock;
    constructor Create;
    destructor Destroy; override;
  end;

function TimeEx: Longint;
function DateTimeToUnix(ConvDate: TDateTime): Longint;
function DateTimeToFileName(DateTime: TDateTime; Extension: string = ''): string;
function StrToHex(S: RawByteString): string;
function HexToStr(S: string): RawByteString;
procedure Explode(Separator: Char; S: string; Strings: TStringList; Limit: Integer = 0);
function SizeToStr(I: Integer): string;
function GetPacketName(Pck: string; FromClient: Boolean): string;

implementation

function DateTimeToFileName(DateTime: TDateTime; Extension: string = ''): string;
begin                    
  Result := FormatDateTime('yyyy-mm-dd-hh-nn', DateTime);
  if Extension <> '' then
    begin
      if Extension[1] <> '.' then Extension := '.' + Extension;
      Result := Result + Extension;
    end;
end;

procedure Explode(Separator: Char; S: string; Strings: TStringList; Limit: Integer = 0);
var
  I, N: Integer;
  Start: Integer;
  Done: Boolean;
begin
  Strings.BeginUpdate;
  try
   Strings.Clear;
   if S = '' then Exit;
   N := 0;
   Start := 1;
   Done := False;
   repeat
     Inc(N);
     I := PosEx(Separator, S, Start);
     if (I = 0) or (N = Limit) then
       begin
         I := Length(S) + 1;
         Done := True;
       end;
     Strings.Add(Copy(S, Start, I - Start));
     Start := I + 1;
   until Done;
  finally
    Strings.EndUpdate;
  end;
end;

function DateTimeToUnix(ConvDate: TDateTime): Longint;
begin
  Result := Round((ConvDate - 25569.0) * 86400);
end;

function TimeEx: Longint;
var
  T: _SystemTime;
begin
  GetSystemTime(T);
  Result := DateTimeToUnix(SystemTimeToDateTime(T));
end;

constructor TSafeStringList.Create;
begin
  inherited Create;
  CriticalSection := TCriticalSection.Create;
end;

destructor TSafeStringList.Destroy;
begin
  CriticalSection.Free;
  inherited Destroy;
end;

procedure TSafeStringList.Lock;
begin
  CriticalSection.Enter;
end;

procedure TSafeStringList.Unlock;
begin
  CriticalSection.Leave;
end;

function HexToStr(S: string): RawByteString;
var
  I, L: Integer;
  B: string;
begin
  B := '';
  S := AnsiLowerCase(S);
  for I := 1 to Length(S) do if AnsiChar(S[I]) in ['a'..'f', '0'..'9'] then B := B + S[I];
  L := Length(B);
  if L mod 2 <> 0 then
    begin
      Exit;
    end;
  Result := '';
  for I := 1 to L div 2 do
    begin
      Result := Result + AnsiChar(StrToInt('$' + Copy(B, I * 2 - 1, 2)));
    end;
end;

function StrToHex(S: RawByteString): string;
var
  I: integer;
begin
  Result := '';
  for I := 1 to Length(S) do
    begin
      if Result <> '' then Result := Result + ' ';
      Result := Result + IntToHex(Ord(S[I]), 2);
    end;
end;

function GetPacketName(Pck: string; FromClient: Boolean): string;
var
  S: string;
  Id: string;
begin
  if Length(Pck) = 0 then
    begin
      Result := 'Empty Packet';
      Exit;
    end;
  Id := IntToHex(Ord(Pck[1]), 2);
  if FromClient then
    begin
      S := 'Unknown';
    end
  else
    begin
      case Ord(Pck[1]) of
        $00: S := 'Die';
        $0C: S := 'AbstractNpcInfo';
        $31: S := 'CharInfo';
        $32: S := 'UserInfo';
        $4A: S := 'CreatureSay';
        else S := 'Unknown';
      end;
    end;
  Result := '[' + Id + '] ' + S;
end;

function SizeToStr(I: Integer): string;
begin
  if I < 1024 * 1000 then
    begin
      Result := IntToStr(Round(I / 1024)) + ' KB';
    end
  else
    begin
      Result := FormatFloat('0.000', Round(I / 1024) / 1000) + ' MB';
    end;
end;

end.
