unit uMain;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, IdContext, IdBaseComponent, IdComponent,
  IdCustomTCPServer, IdTCPServer, GameStream, IdIOHandler, IdGlobal, SyncObjs, Basic,
  Vcl.StdCtrls, IniFiles, Registry, Items, Vcl.ComCtrls, Vcl.ExtCtrls, ShellApi, MD5;

const
  WM_PRISM = WM_USER + 5123;
  PM_LOGGER_EVENT = 1;
  RACE_NUIAN = 1;
  RACE_ELF = 4;
  RACE_HARIHARAN = 5;
  RACE_FERRE = 6;

type
  TPrismEventType = (etDefault, etError, etUser, etNetwork, etImportant);

  TCharacter = record
    Id: Integer;
    Name: string;
    Level: Integer;
    Weapon, WeaponExtra, WeaponRanged, Instrument: Integer;
    Head, Chest, Legs, Feet, Gloves: Integer;

    Model: TCharacterModel;
  end;

  TPrismLogger = class(TThread)
  private
    FFormHandle: THandle;
    FList: TSafeStringList;

    FEventText: string;
    FEventType: TPrismEventType;

  protected
    procedure Execute; override;

  public
    property FormHandle: THandle read FFormHandle write FFormHandle;
    property EventText: string read FEventText;
    property EventType: TPrismEventType read FEventType;

    procedure Add(Text: string; EventType: TPrismEventType = etDefault; Debug: Boolean = False);

    constructor Create;
    destructor Destroy; override;
  end;

  TGameLauncher = class(TThread)
  private
    FFilename, FParam: string;

  protected
    procedure Execute; override;

  public
    constructor Create(Filename, Param: string);
    destructor Destroy; override;
  end;

  TfrmMain = class(TForm)
    LoginServer: TIdTCPServer;
    GameServer: TIdTCPServer;
    UnknownServer: TIdTCPServer;
    btnRun: TButton;
    OpenDialog: TOpenDialog;
    cbRace: TComboBox;
    cbGender: TComboBox;
    lRace: TLabel;
    lGender: TLabel;
    PageControl: TPageControl;
    : TTabSheet;
    : TTabSheet;
    eFilter: TEdit;
    lbSets: TListBox;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    Label5: TLabel;
    Label6: TLabel;
    cbLegs: TComboBox;
    cbChest: TComboBox;
    cbGloves: TComboBox;
    cbFeet: TComboBox;
    cbHead: TComboBox;
    cbWeapon: TComboBox;
    LinkLabel1: TLinkLabel;
    LinkLabel2: TLinkLabel;
    LinkLabel3: TLinkLabel;
    procedure FormCreate(Sender: TObject);
    procedure GameServerExecute(AContext: TIdContext);
    procedure LoginServerExecute(AContext: TIdContext);
    function NewStream(Tag: Integer; Master: TObject): TGameStream;
    procedure UnknownServerExecute(AContext: TIdContext);
    procedure btnRunClick(Sender: TObject);
    procedure eFilterChange(Sender: TObject);
    procedure lbSetsClick(Sender: TObject);
    procedure cbRaceChange(Sender: TObject);
    procedure cbGenderChange(Sender: TObject);
    procedure LoadItems;
    procedure cbItemChange(Sender: TObject);
    procedure LinkLabel1LinkClick(Sender: TObject; const Link: string;
      LinkType: TSysLinkType);
    procedure LinkLabel2LinkClick(Sender: TObject; const Link: string;
      LinkType: TSysLinkType);
    procedure LinkLabel3LinkClick(Sender: TObject; const Link: string;
      LinkType: TSysLinkType);
  private
    FLogger: TPrismLogger;
    Player: TCharacter;
    Path: string;
    TaskIndex: Integer;
    LastSet: Integer;

    procedure FillSets;
    procedure FillModels;
    procedure SelectNewSet;
    procedure ProcessModel;
    procedure ProcessItems(Sender: TObject);

    procedure SendCharacterList(Stream: TGameStream; Char: TCharacter);
    procedure SendDeleteCharacterResponse(Stream: TGameStream; CharId: Integer);

    procedure PrismMessage(var Msg: TMessage); message WM_PRISM;
    function IsValidMD5: Boolean;
  public
    procedure PacketEvent(Packet: TGamePacket; Sender: TGameStream);
    procedure SendEvent(Data: RawByteString; Sender: TGameStream);
  end;

var
  frmMain: TfrmMain;

const
  ProjectName = 'AAplay.ru';

  LSPacket1: RawByteString = RawByteString(#$00#$00 +
  #$00#$00#$06#$00#$00#$00#$00#$00#$00#$00);

  LSPacket2: RawByteString = RawByteString(#$03#$00 +
  #$58#$33#$00#$00#$20#$00#$32#$36#$39#$36#$31#$32#$65#$37#$61#$36 +
  #$30#$39#$34#$31#$31#$38#$62#$62#$37#$35#$30#$37#$64#$62#$63#$34 +
  #$32#$62#$61#$35#$39#$34);

  GSPacket1: RawByteString = RawByteString(#$DD#$01#$00#$00 +
  #$00#$00#$00#$95#$FF#$3C#$1E#$E2#$04#$22#$26#$EE#$52#$00#$00#$00 +
  #$00);

  GSPacket4: RawByteString = RawByteString(#$DD#$02#$0F#$00 +
  #$13#$00#$65#$5F#$66#$61#$6C#$63#$6F#$6E#$79#$5F#$70#$6C#$61#$74 +
  #$65#$61#$75#$5F#$31#$00#$00#$00#$00#$00#$00#$00#$00#$01);

  GSPacket5: RawByteString = RawByteString(#$DD#$01#$06#$00 +
  #$0A#$00#$61#$61#$2E#$6D#$61#$69#$6C#$2E#$72#$75#$07#$00#$3E#$28 +
  #$0F#$0F#$79#$00#$13#$00#$00#$00#$00#$00#$0A#$00#$32#$00#$00#$00 +
  #$00#$00#$01#$01#$01#$01#$01);

  GSPacket6: RawByteString = RawByteString(#$DD#$01#$B9#$01 +
  #$01#$00#$00#$00#$01#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00 +
  #$12#$B3#$15#$53#$00#$00#$00#$00);

  GSPacket7: RawByteString = RawByteString(#$DD#$01#$CC#$00 +
  #$00#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00);

  GSPacket10: RawByteString = RawByteString(#$DD#$01#$B9#$01 +
  #$05#$00#$00#$00#$01#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00 +
  #$12#$B3#$15#$53#$00#$00#$00#$00);

  GSPacket11: RawByteString = RawByteString(#$DD#$01#$C1#$01 +
  #$B8#$5C#$01#$00#$08#$00#$D0#$97#$D1#$8D#$D0#$BD#$D0#$B0#$10#$00 +
  #$63#$68#$61#$72#$61#$63#$74#$65#$72#$5F#$6F#$70#$74#$69#$6F#$6E +
  #$00#$00#$00#$00#$00#$00);

  GSPacket12: RawByteString = RawByteString(#$DD#$01#$C1#$01 +
  #$B8#$5C#$01#$00#$08#$00#$D0#$97#$D1#$8D#$D0#$BD#$D0#$B0#$0B#$00 +
  #$6B#$65#$79#$5F#$62#$69#$6E#$64#$69#$6E#$67#$00#$00#$00#$00#$00 +
  #$00);

  GSPacket13: RawByteString = RawByteString(#$DD#$01#$3B#$00 +
  #$00#$00#$00#$00#$00#$00#$00#$00#$00);

implementation

{$R *.dfm}

constructor TGameLauncher.Create(Filename, Param: string);
begin
  FreeOnTerminate := True;
  inherited Create(False);
  FFilename := Filename;
  FParam := Param;
end;

destructor TGameLauncher.Destroy;
begin
  inherited Destroy;
end;

procedure TGameLauncher.Execute;
begin
  WinExec(PAnsiChar(AnsiString(FFilename + ' ' + FParam)), SW_NORMAL);
end;

constructor TPrismLogger.Create;
begin
  FreeOnTerminate := True;
  FList := TSafeStringList.Create;

  inherited Create(True);
end;

destructor TPrismLogger.Destroy;
begin
  FList.Free;

  inherited Destroy;
end;

procedure TPrismLogger.Execute;
var
  SL: TStringList;
  I: Integer;
begin
  SL := TStringList.Create;
  try
    while not Terminated do
      begin
        SL.Clear;
        FList.Lock;
        try
          SL.AddStrings(FList);
          FList.Clear;
        finally
          FList.Unlock;
        end;

        if SL.Count = 0 then
          begin
            Sleep(100);
          end
        else
          begin
            for I := 0 to SL.Count - 1 do
              begin
                if FFormHandle <> 0 then
                  begin
                    FEventType := TPrismEventType(SL.Objects[I]);
                    FEventText := SL[I];
                    SendMessage(FFormHandle, WM_PRISM, PM_LOGGER_EVENT, Integer(Self));
                  end;
              end;
          end;
      end;
  finally
    SL.Free;
  end;
end;

procedure TPrismLogger.Add(Text: string; EventType: TPrismEventType = etDefault; Debug: Boolean = False);
begin
  FList.Lock;
  try
    FList.AddObject(Text, TObject(EventType));
  finally
    FList.Unlock;
  end;
end;

procedure TfrmMain.PrismMessage(var Msg: TMessage);
var
  Logger: TPrismLogger;
begin
  case Msg.WParam of
    PM_LOGGER_EVENT:
      begin
        Logger := TPrismLogger(Msg.LParam);
        if Logger.FEventType = etImportant then ShowMessage(Logger.EventText);
      end;
  end;
end;

procedure TfrmMain.LinkLabel1LinkClick(Sender: TObject; const Link: string;
  LinkType: TSysLinkType);
begin
  ShellExecute(0, '', PChar(Link), '', '', SW_SHOW);
end;

procedure TfrmMain.LinkLabel2LinkClick(Sender: TObject; const Link: string;
  LinkType: TSysLinkType);
begin
  ShellExecute(0, '', PChar(Link), '', '', SW_SHOW);
end;

procedure TfrmMain.LinkLabel3LinkClick(Sender: TObject; const Link: string;
  LinkType: TSysLinkType);
begin
  ShellExecute(0, '', PChar(Link), '', '', SW_SHOW);
end;

procedure TfrmMain.LoadItems;
var
  A, SL: TStringList;
  Filename: string;
  I, Id, Slot: Integer;
  Name: string;
  CB: TComboBox;
  List: array of TComboBox;
const
  DataFile = 'Items.txt';
begin
  Filename := Path + DataFile;
  if not FileExists(Filename) then
    begin
      Application.MessageBox(PChar(Format('     (%s).', [DataFile])), ProjectName, MB_ICONERROR);
      Exit;
    end;
  SetLength(List, 6);
  List[0] := cbHead;
  List[1] := cbChest;
  List[2] := cbLegs;
  List[3] := cbFeet;
  List[4] := cbGloves;
  List[5] := cbWeapon;

  for I := 0 to High(List) do List[I].Items.Clear;

  SL := TStringList.Create;
  A := TStringList.Create;
  try
    SL.LoadFromFile(Filename);
    for I := 0 to High(List) do List[I].Items.BeginUpdate;
    try
      for I := 0 to High(List) do List[I].Items.AddObject('', TObject(0));
      for I := 0 to SL.Count - 1 do
        begin
          Explode(#9, SL[I], A, 3);
          if A.Count <> 3 then Continue;
          Id := StrToIntDef(A[0], 0);
          Slot := StrToIntDef(A[1], 0);
          Name := A[2];
          if (Id < 1) or (Slot = 0) or (Name = '') then Continue;
          case Slot of
            -1: CB := cbWeapon;
            1: CB := cbHead;
            3: CB := cbChest;
            5: CB := cbLegs;
            6: CB := cbGloves;
            7: CB := cbFeet;
            else CB := nil;
          end;
          if CB = nil then Continue;
          CB.Items.AddObject(Name, TObject(Id));
        end;
    finally
      for I := 0 to High(List) do List[I].Items.EndUpdate;
    end;
  finally
    FreeAndNil(SL);
    FreeAndNil(A);
  end;

  for I := 0 to High(List) do List[I].Sorted := True;
end;

function GetArcheAgeFilename: string;
var
  R: TRegistry;
  S: string;
  Ini: TIniFile;
  Path: string;
begin
  Result := '';
  try
    S := '';
    R := TRegistry.Create;
    try
      R.RootKey := HKEY_CURRENT_USER;
      if not R.OpenKey('Software\Mail.Ru\GameCenterMailRu', False) then Exit;
      S := R.ReadString('1');
    finally
      R.Free;
    end;

    if S = '' then Exit;

    S := ExtractFilePath(S) + 'GameCenter@Mail.Ru.ini';
    if not FileExists(S) then Exit;

    Path := '';
    Ini := TIniFile.Create(S);
    try
      Path := Ini.ReadString('Main', 'GamesInstallPath', '');
    finally
      Ini.Free;
    end;

    if Path = '' then Exit;
    S := IncludeTrailingBackSlash(Path) + 'ArcheAge\Bin32\archeage.exe';
    if not FileExists(S) then Exit;

    Result := S;
  except
    // nothing
  end;
end;

procedure TfrmMain.btnRunClick(Sender: TObject);
var
  Ini: TIniFile;
  Filename, Param: string;
begin
  Param := '+acpxmk -r +auth_ip 127.0.0.1:1237 -uid aaplay -token 1111111111111111111111111111111111111111';
  Filename := '';

  Ini := TIniFile.Create(Path + 'Options.ini');
  try
    Filename := Ini.ReadString('ArcheAge', 'Filename', '');
    if Filename <> '' then
      begin
        if not FileExists(Filename) then
          begin
            Filename := '';
          end;
      end;
    if Filename = '' then
      begin
        Filename := GetArcheageFilename;
        if Filename = '' then
          begin
            if OpenDialog.Execute then
              begin
                Filename := OpenDialog.FileName;
              end;
          end;
        if Filename <> '' then
          begin
            Ini.WriteString('ArcheAge', 'Filename', Filename);
          end;
      end;
  finally
    Ini.Free;
  end;

  if Filename <> '' then
    begin
      TGameLauncher.Create(Filename, Param);
    end;
end;

procedure TfrmMain.ProcessModel;
var
  I, Race, Gender: Integer;
begin
  Race := Integer(cbRace.Items.Objects[cbRace.ItemIndex]);
  Gender := Integer(cbGender.Items.Objects[cbGender.ItemIndex]);
  for I := 0 to High(CharacterModels) do
    begin
      if (CharacterModels[I].race = Race) and (CharacterModels[I].gender = Gender) then
        begin
          Player.Model := CharacterModels[I];
          Inc(TaskIndex);
          Exit;
        end;
    end;
end;

procedure TfrmMain.cbGenderChange(Sender: TObject);
begin
  ProcessModel;
end;

procedure TfrmMain.ProcessItems(Sender: TObject);
var
  CB: TComboBox;
  Id: Integer;
  List: array of TComboBox;
  I: Integer;
begin
  SetLength(List, 6);
  List[0] := cbHead;
  List[1] := cbChest;
  List[2] := cbLegs;
  List[3] := cbFeet;
  List[4] := cbGloves;
  List[5] := cbWeapon;

  for I := 0 to High(List) do
    begin
      CB := List[I];
      if (CB.ItemIndex >= 0) and (CB.ItemIndex < CB.Items.Count) then
        begin
          Id := Integer(CB.Items.Objects[CB.ItemIndex])
        end
      else Id := 0;

      case I of
        0: Player.Head := Id;
        1: Player.Chest := Id;
        2: Player.Legs := Id;
        3: Player.Feet := Id;
        4: Player.Gloves := Id;
        5: Player.Weapon := Id;
      end;
    end;

  if Sender <> cbWeapon then
    begin
      LastSet := -1;
      lbSets.ItemIndex := -1;
    end;
  Inc(TaskIndex);
end;

procedure TfrmMain.cbItemChange(Sender: TObject);
begin
  if lbSets.Tag <> 1 then ProcessItems(Sender);
end;

procedure TfrmMain.cbRaceChange(Sender: TObject);
begin
  ProcessModel;
end;

procedure TfrmMain.eFilterChange(Sender: TObject);
begin
  FillSets;
end;

procedure TfrmMain.SelectNewSet;
var
  I: Integer;
  S: TItemSet;
begin
  I := lbSets.ItemIndex;
  if (I < 0) or (I > lbSets.Items.Count - 1) then Exit;
  I := Integer(lbSets.Items.Objects[I]);
  if I = LastSet then Exit;
  LastSet := I;

  S := ItemSets[I];

  Player.Head := S.Head;
  Player.Chest := S.Chest;
  Player.Legs := S.Legs;
  Player.Gloves := S.Gloves;
  Player.Feet := S.Feet;
  Inc(TaskIndex);

  lbSets.Tag := 1;
  try
    cbHead.ItemIndex := cbHead.Items.IndexOfObject(TObject(Player.Head));
    cbChest.ItemIndex := cbChest.Items.IndexOfObject(TObject(Player.Chest));
    cbLegs.ItemIndex := cbLegs.Items.IndexOfObject(TObject(Player.Legs));
    cbGloves.ItemIndex := cbGloves.Items.IndexOfObject(TObject(Player.Gloves));
    cbFeet.ItemIndex := cbFeet.Items.IndexOfObject(TObject(Player.Feet));
  finally
    lbSets.Tag := 0;
  end;
end;

procedure TfrmMain.FillModels;
begin
  cbRace.Items.Clear;
  cbRace.Items.AddObject('', TObject(RACE_NUIAN));
  cbRace.Items.AddObject('', TObject(RACE_ELF));
  cbRace.Items.AddObject('', TObject(RACE_HARIHARAN));
  cbRace.Items.AddObject('', TObject(RACE_FERRE));

  cbGender.Items.Clear;
  cbGender.Items.AddObject('', TObject(1));
  cbGender.Items.AddObject('', TObject(2));

  cbRace.ItemIndex := 0;
  cbGender.ItemIndex := 0;
end;

procedure TfrmMain.FillSets;
var
  I: Integer;
  S: string;
  Filter: Boolean;
begin
  S := AnsiLowerCase(Trim(eFilter.Text));
  Filter := eFilter.Text <> '';

  lbSets.ItemIndex := -1;
  lbSets.Items.BeginUpdate;
  lbSets.Items.Clear;
  for I := 0 to High(ItemSets) do
    begin
      if Filter then
        begin
          if Pos(S, AnsiLowerCase(ItemSets[I].Name)) < 1 then Continue;
        end;
      lbSets.Items.AddObject(ItemSets[I].Name, TObject(I));
    end;
  lbSets.Items.EndUpdate;

  if lbSets.Items.Count = 1 then
    begin
      lbSets.ItemIndex := 0;
    end;
  SelectNewSet;
end;

procedure TfrmMain.FormCreate(Sender: TObject);
begin
  if not IsValidMD5 then Halt;

  Path := ExtractFilePath(ParamStr(0));
  TaskIndex := 0;

  LoginServer.DefaultPort := 1237;
  GameServer.DefaultPort := 1239;
  UnknownServer.DefaultPort := 1250;
  LoginServer.Active := True;
  GameServer.Active := True;
  UnknownServer.Active := True;

  Player.Id := 1;
  Player.Name := ProjectName;
  Player.Level := 1;

  Player.Weapon := 0;
  Player.WeaponExtra := 0;
  Player.WeaponRanged := 0;
  Player.Instrument := 0;

  Player.Head := 0;
  Player.Chest := 0;
  Player.Legs := 0;
  Player.Feet := 0;
  Player.Gloves := 0;

  FLogger := TPrismLogger.Create;
  FLogger.FormHandle := Handle;
  FLogger.Start;

  LastSet := -1;

  FillSets;
  FillModels;
  ProcessModel;
  LoadItems;
end;

function TfrmMain.NewStream(Tag: Integer; Master: TObject): TGameStream;
var
  S: TGameStream;
begin
  S := TGameStream.Create;
  S.Tag := Tag;
  S.Master := Master;
  S.PacketEvent := PacketEvent;
  S.SendEvent := SendEvent;

  Result := S;
end;

procedure TfrmMain.SendDeleteCharacterResponse(Stream: TGameStream; CharId: Integer);
var
  P: TGamePacket;
begin
  P := TGamePacket.Create;
  try
    P.Id := $003501DD;
    P.WriteD(CharId);
    P.WriteC(1);
    P.WriteQ(0);
    P.WriteQ(0);
    Stream.Send(P);
  finally
    FreeAndNil(P);
  end;
end;

procedure TfrmMain.SendCharacterList(Stream: TGameStream; Char: TCharacter);
var
  P: TGamePacket;
  S: RawByteString;
  I: Integer;

  procedure WriteItem(ItemId: Integer);
  var
    I: Integer;
  begin
    P.WriteD(ItemId);
    if ItemId > 0 then
      begin
        P.WriteD(1); // ObjectId
        for I := 1 to 6 do P.WriteC(0);
        P.WriteC(1);
        for I := 1 to 3 do P.WriteC(0);
        P.WriteC(1);
        for I := 1 to 4 do P.WriteC(0);
        P.WriteC(1);
        for I := 1 to 62 do P.WriteC(0);
        P.WriteC(3);
        for I := 1 to 16 do P.WriteC(0);
      end;
  end;
begin
  P := TGamePacket.Create;
  try
    P.Id := $003A01DD;
    P.WriteC(1);
    P.WriteC(1);
    P.WriteD(Char.Id);
    P.WriteS(Char.Name);

    P.WriteC(Char.Model.race);
    P.WriteC(Char.Model.gender);
    P.WriteC(Char.Level);
    P.WriteC($64);
    P.WriteC(1);
    P.WriteC(0);
    P.WriteC(0);
    P.WriteD($178);
    P.WriteD(180); // zone_id?
    P.WriteC(101); // faction_id?
    for I := 1 to 21 do P.WriteC(0);

    WriteItem(Char.Chest);
    WriteItem(Char.Head);
    WriteItem(Char.Legs);
    WriteItem(Char.Gloves);
    WriteItem(Char.Feet);

    P.WriteD(0);
    P.WriteD(0);
    P.WriteD(0);
    P.WriteD(0);
    P.WriteD(0);
    P.WriteD(0);
    P.WriteD(0);
    P.WriteD(0);

    WriteItem(Char.Weapon);
    WriteItem(Char.WeaponExtra);
    WriteItem(Char.WeaponRanged);
    WriteItem(Char.Instrument);

    P.WriteD(Char.Model.face);
    P.WriteD(Char.Model.hair_id);

    P.WriteD(0);
    P.WriteD(0);
    P.WriteD(0);

    P.WriteD(Char.Model.body);

    S := RawByteString(#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00#$01 +
    #$0B#$0B#$00#$00#$00#$00#$28#$5F#$CD#$03#$00#$00#$00#$00#$40#$2F +
    #$C1#$03#$AC#$7D#$FC#$42#$03);
    P.WriteB(@S[1], Length(S));

    P.WriteD(Char.Model.hair_color_id);
    P.WriteD(Char.Model.skin_color_id);
    P.WriteD(0);
    P.WriteD(0);
    P.WriteSingle(1);
    P.WriteSingle(1);
    P.WriteD(0);
    P.WriteD(0);
    P.WriteD(Char.Model.face_fixed_decal_asset_0_id);
    P.WriteSingle(Char.Model.face_fixed_decal_asset_0_weight);
    P.WriteD(Char.Model.face_fixed_decal_asset_1_id);
    P.WriteSingle(Char.Model.face_fixed_decal_asset_1_weight);
    P.WriteD(Char.Model.face_fixed_decal_asset_2_id);
    P.WriteSingle(Char.Model.face_fixed_decal_asset_2_weight);
    P.WriteD(Char.Model.face_fixed_decal_asset_3_id);
    P.WriteSingle(Char.Model.face_fixed_decal_asset_3_weight);
    P.WriteD(0);
    P.WriteD(Char.Model.face_normal_map_id);
    P.WriteD(0);
    P.WriteSingle(1);
    P.WriteD(0);
    P.WriteD(Char.Model.left_pupil_color);
    P.WriteD(Char.Model.right_pupil_color);
    P.WriteD(Char.Model.eyebrow_color);
    P.WriteD(Char.Model.deco_color);

    S := HexToStr(Char.Model.modifier);
    P.WriteH(Length(S));
    if S <> '' then
      begin
        P.WriteB(@S[1], Length(S));
      end;

    P.WriteD(500); //  

    S := RawByteString(#$BF +
    #$F5#$F0#$52#$00#$00#$00#$00#$00#$00#$6B#$52#$EF#$52#$00#$00#$00 +
    #$00#$00#$00#$00#$00#$6B#$52#$EF#$52#$00#$00#$00#$00#$00#$00#$00 +
    #$00#$7A#$F5#$F0#$52#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00 +
    #$00#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00 +
    #$00#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00 +
    #$00#$00#$00#$00#$00#$00#$00#$00#$00#$03#$00#$00#$00#$00#$00#$00 +
    #$00#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00 +
    #$00#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00#$00 +
    #$00#$00#$5E#$D5#$F0#$52#$00#$00#$00#$00);
    P.WriteB(@S[1], Length(S));

    Stream.Send(P);
  finally
    FreeAndNil(P);
  end;
end;

procedure TfrmMain.PacketEvent(Packet: TGamePacket; Sender: TGameStream);
var
  S: RawByteString;
  P: TGamePacket;
  I: Integer;
  PacketId: Cardinal;
  Stream: TGameStream;
  Ping: array [0..4] of Integer;
begin
  Stream := Sender;
  P := TGamePacket.Create;
  try
    S := Packet.Data;
    if Stream.Tag = 0 then
      begin
        PacketId := Packet.Id and $0000FFFF;
        case PacketId of
          $04, $0D:
            begin
              Stream.Send(LSPacket1);
              Stream.Send(LSPacket2);
            end;

          $0A:
            begin
              P.Id := $08;

              P.WriteC(1); // Count

              P.WriteC(1);
              P.WriteS(ProjectName);
              P.WriteC(1);
              P.WriteH(0);
              P.WriteH(0);
              P.WriteD(0);
              P.WriteH(0);

              P.WriteC(1); // Count

              P.WriteD(1);
              P.WriteC(1);
              P.WriteD(1);
              P.WriteS(ProjectName);
              P.WriteC(1);
              P.WriteC(1);
              P.WriteS('');
              P.WriteD(0);
              P.WriteD(0);

              S := P.Data;
              Delete(S, 3, 2);

              Stream.Send(S);
            end;

          $0B:
            begin
              S := RawByteString('127.0.0.1');
              P.Data := RawByteString(#$0A#$00#$3D#$1A#$15#$E8);
              I := Length(S);
              P.WriteH(I);
              P.WriteB(@S[1], I);
              P.WriteH(GameServer.DefaultPort);
              Stream.Send(P);
            end;
        end;
      end;

    if Stream.Tag = 1 then
      begin
        PacketId := Packet.Id and $FFFFFF00;
        //if PacketId <> $00120200 then FLogger.Add(StrToHex(Packet.Data));

        case PacketId of
          $00120200:
            begin
              for I := 0 to 4 do
                begin
                  Ping[I] := Packet.ReadD;
                end;

              P.Id := $001302DD;
              for I := 0 to 3 do
                begin
                  P.WriteD(Ping[I]);
                end;

              I := GetTickCount;
              P.WriteD(0);
              P.WriteD(0);
              P.WriteQ(I * 1000);
              P.WriteD(Ping[4]);
              P.WriteD(I);

              Stream.Send(P);
            end;

          $00000100:
            begin
              Stream.Send(GSPacket1);

              P.Id := $000002DD;
              P.WriteD(0);
              Stream.Send(P);
            end;

          $00010200:
            begin
              I := Packet.ReadD;

              if I = 0 then
                begin
                  P.Id := $000002DD;
                  P.WriteD(1);
                  Stream.Send(P);

                  Stream.Send(GSPacket4);
                  Stream.Send(GSPacket5);
                  Stream.Send(GSPacket6);
                  Stream.Send(GSPacket7);
                end
              else
                begin
                  P.Id := $000002DD;
                  P.WriteD(2);
                  Stream.Send(P);
                end;
            end;

          $001F0100:
            begin
              Stream.Send(GSPacket10);
              Stream.Send(GSPacket11);
              Stream.Send(GSPacket12);
              Stream.Send(GSPacket13);

              SendCharacterList(Stream, Player);

              P.Id := $019801DD;
              P.WriteC(1);
              P.WriteS('    ' + ProjectName + '!');
              Stream.Send(P);

              Stream.Game := True;
            end;

          $00010100:
            begin
              P.Id := $000101DD;
              P.WriteD(1);
              Stream.Send(P);
            end;

          $00230100:
            begin
              I := Packet.ReadD;
              SendDeleteCharacterResponse(Stream, I);

              SendCharacterList(Stream, Player);
            end;

          else
            begin
              FLogger.Add('Unknown: ' + IntToHex(PacketId, 8));
            end;
        end;
      end;

    if Stream.Tag = 2 then
      begin
        Stream.Send(#1#0#0);
      end;
  finally
    FreeAndNil(P);
  end;
end;

procedure TfrmMain.SendEvent(Data: RawByteString; Sender: TGameStream);
var
  C: TIdContext;
  B: TIdBytes;
  L: Integer;
begin
  L := Length(Data);
  SetLength(B, L);
  Move(Data[1], B[0], L);
  C := TIdContext(Sender.Master);
  C.Connection.IOHandler.Write(B);
end;

procedure TfrmMain.UnknownServerExecute(AContext: TIdContext);
var
  Stream: TGameStream;
  IO: TIdIOHandler;
  B: TIdBytes;
  S: RawByteString;
  L: Integer;
begin
  try
    Stream := NewStream(2, AContext);
    IO := AContext.Connection.IOHandler;
    try
      while AContext.Connection.Connected do
        begin
          IO.CheckForDataOnSource(10);
          if not IO.InputBufferIsEmpty then
            begin
              SetLength(B, 0);
              IO.InputBuffer.ExtractToBytes(B);
              IO.InputBuffer.Clear;

              L := Length(B);
              SetLength(S, L);
              Move(B[0], S[1], L);

              Stream.ProcessData(S);
            end;
        end;
    finally
      FreeAndNil(Stream);
    end;
  except
    // nothing
  end;
end;

procedure TfrmMain.GameServerExecute(AContext: TIdContext);
var
  Stream: TGameStream;
  IO: TIdIOHandler;
  B: TIdBytes;
  S: RawByteString;
  L: Integer;
  TI: Integer;

begin
  TI := 0;
  try
    FLogger.Add('Game Execute');
    Stream := NewStream(1, AContext);
    Stream.Game := False;
    IO := AContext.Connection.IOHandler;
    try
      while AContext.Connection.Connected do
        begin
          IO.CheckForDataOnSource(10);
          if not IO.InputBufferIsEmpty then
            begin
              SetLength(B, 0);
              IO.InputBuffer.ExtractToBytes(B);
              IO.InputBuffer.Clear;

              L := Length(B);
              SetLength(S, L);
              Move(B[0], S[1], L);

              Stream.ProcessData(S);

              if Stream.Game then
                begin
                  if TaskIndex > TI then
                    begin
                      TI := TaskIndex;
                      SendDeleteCharacterResponse(Stream, Player.Id);
                      SendCharacterList(Stream, Player);

                      FLogger.Add('Done!');
                    end;
                end;
            end;
        end;
    finally
      FreeAndNil(Stream);
      FLogger.Add('Game Halt');
    end;
  except
    // nothing
  end;
end;

procedure TfrmMain.lbSetsClick(Sender: TObject);
begin
  SelectNewSet;
end;

procedure TfrmMain.LoginServerExecute(AContext: TIdContext);
var
  Stream: TGameStream;
  IO: TIdIOHandler;
  B: TIdBytes;
  S: RawByteString;
  L: Integer;
begin
  try
    FLogger.Add('Login Execute');
    Stream := NewStream(0, AContext);
    IO := AContext.Connection.IOHandler;
    try
      while AContext.Connection.Connected do
        begin
          IO.CheckForDataOnSource(10);
          if not IO.InputBufferIsEmpty then
            begin
              SetLength(B, 0);
              IO.InputBuffer.ExtractToBytes(B);
              IO.InputBuffer.Clear;

              L := Length(B);
              SetLength(S, L);
              Move(B[0], S[1], L);

              Stream.ProcessData(S);
            end;
        end;
    finally
      FreeAndNil(Stream);
      FLogger.Add('Login Halt');
    end;
  except
    // nothing
  end;
end;

function TfrmMain.IsValidMD5: Boolean;
var
  F: TFileStream;
  S: RawByteString;
  L: Integer;
  MD5: string;
begin
  F := TFileStream.Create(ParamStr(0), fmOpenRead or fmShareDenyWrite);
  try
    L := F.Size - 16;
    SetLength(S, L);
    F.Read(S[1], L);
    MD5 := GetMD5(S);
    SetLength(S, 16);
    F.Read(S[1], 16);
    Result := HexToStr(MD5) = S;
  finally
    FreeAndNil(F);
  end;
end;

end.
